#include "BlueNoise.hlsl"
#include "Engine/Common.hlsl"

struct FragmentNode
{
  uint Next;
  float Depth;
};

struct FragmentData
{
  uint Color;
  uint Emissive;
  uint Specular;
  uint Normal;
};

cbuffer TransparencyParameters
{
  uint NodeAlloc;
  uint Width;
  uint WidthDivisor;
  uint HeightDivisor;
  float beersScaleFactor;
};

uint PackColor(float4 color, int2 ScreenPos)
{
  uint result = 0;
  float a = color.a;
  /* float4 dither = (blueNoise4(ScreenPos)*2.0-1.0)*(1.0/255.0); */
  /* color = saturate(color+dither); */
  color.a = (a == 0.0) ? 0.0 : color.a;
  result |= uint(color.r * 255.0);
  result |= uint(color.g * 255.0) << 8;
  result |= uint(color.b * 255.0) << 16;
  result |= uint(color.a * 255.0) << 24;
  return result;
}

uint PackColor(float3 color, int2 ScreenPos)
{
  uint result = 0;
  /* float3 dither = (blueNoise3(ScreenPos)*2.0-1.0)*(1.0/255.0); */
  /* color = saturate(color+dither); */
  result |= uint(color.r * 255.0);
  result |= uint(color.g * 255.0) << 8;
  result |= uint(color.b * 255.0) << 16;
  return result;
}

float4 UnpackColor(uint color)
{
  float4 result = float4(0, 0, 0, 0);

  result.r = (color & 0x000000FF);
  result.g = (color & 0x0000FF00) >> 8;
  result.b = (color & 0x00FF0000) >> 16;
  result.a = (color & 0xFF000000) >> 24;

  return result / 255.0;
}

uint PackNormal(float3 normal)
{
  // Use spheremap transform, we let positive Z be in the center where
  // there is most precision, and -Z have the least.
  float2 n = float2(0, 0);

  if (normal.x != 0 || normal.y != 0) {
    n = normalize(normal.xy) * sqrt(-normal.z * 0.5 + 0.5);
  }

  uint2 i = f32tof16(n);
  return i.x | (i.y << 16u);
}

float3 UnpackNormal(uint n)
{
  float2 normal = f16tof32(uint2(n & 0xffffu, n >> 16u));
  float3 result = float3(0, 0, 0);

  float s = dot(normal.xy, normal.xy);
  result.z = -2.0 * (s - 0.5);

  // result.xy = normalize(normal.xy)*sqrt(1-result.z^2)
  //           = 1/sqrt(s)*normal.xy*sqrt(1-result.z^2)
  //           = sqrt( (1-result.z^2)/s )*normal.xy
  //           = sqrt( (1-(-2*(s - 0.5))^2)/s )*normal.xy
  //           = sqrt( (1 - 4( s^2 - 2/2*s + 1/4) ) )/s )*normal.xy
  //           = sqrt( (1 - 4s^2 + 4*s - 1 )/s )*normal.xy
  //           = sqrt( (4s - 4s^2)/s )*normal.xy
  //           = 2 sqrt(1 - s)*normal.xy
  result.xy = 2.0 * sqrt(1 - s) * normal.xy;

  return result;
}

void pack(out FragmentData data,
          in float4 ScreenPos,
          in float3 Normal,
          in int LightModel,
          in float Alpha,
          in float3 Emissive,
          in float3 Diffuse,
          in float3 Specular,
          in float SpecularPower)
{
  data.Color = PackColor(float4(Diffuse, Alpha), (int2)ScreenPos.xy);
  data.Emissive = PackColor(Emissive, (int2)ScreenPos.xy) | (LightModel << 24);
  data.Specular = PackColor(float3(Specular), (int2)ScreenPos.xy) | (int(2 * SpecularPower) << 24);
  data.Normal = PackNormal(normalize(Normal));
}

void unpack(out float3 Normal,
            out int LightModel,
            out float Alpha,
            out float3 Emissive,
            out float3 Diffuse,
            out float3 Specular,
            out float SpecularPower,
            in FragmentData data)
{
  Normal = UnpackNormal(data.Normal);
  LightModel = data.Emissive >> 24;
  Alpha = UnpackColor(data.Color).a;
  Emissive = UnpackColor(data.Emissive).rgb;
  Diffuse = UnpackColor(data.Color).rgb;
  Specular = UnpackColor(data.Specular).rgb;
  SpecularPower = 0.5*(data.Specular >> 24);
}
